from __future__ import annotations

import itertools

class Piece:
    """TicTacToe pieces."""

    piece_symbols: str = ".XOVTSZLC"
    piece_unicode_symbols: str = "🟫⬜⬛🟥🟩🟦🟨🟧🟪"
    
    def __init__(self, code: int = 0) -> None:
        """Initializes a Piece instance from its number."""
        self.code: int = code

    def from_symbol(symbol: str) -> Piece:
        """Initializes a Piece instance from its symbol."""
        return Piece(Piece.piece_symbols.index(symbol))

    def next(self, players: int) -> Piece:
        """Returns the piece that will be placed on the next move."""
        if self.code == players:
            return Piece(1)
        else:
            return Piece(self.code+1)

    def previous(self, players: int) -> Piece:
        """Return the piece that was placed on the previous move."""
        if self.code == 1:
            return Piece(players)
        else:
            return Piece(self.code-1)

    def __eq__(self, other: Piece) -> bool:
        return self.code == other.code

    def __int__(self) -> int:
        return self.code

    def __repr__(self) -> str:
        return Piece.piece_symbols[self.code]

    def unicode_symbol(self) -> str:
        return Piece.piece_unicode_symbols[self.code]

class Board:
    """A board representing the position of pieces."""

    def __init__(self, m: int = 3, n: int = 3, infinite: bool = False) -> None:
        """Initializes a board."""
        if m <= 0 or n <= 0:
            raise ValueError("Parameters m, n must be positive.")
        self.infinite: bool = infinite
        self.m: bool = m
        self.n: bool = n
        self.board: list[list[Piece]] = [[Piece()]*m for _ in range(n)]

    def columns(self) -> list[list[Piece]]:
        """Returns columns for checking win."""
        board = self.board
        return [list(square) for square in zip(*board)]

    def diagonals(self) -> list[list[Piece]]:
        """Returns diagonals for checking win."""
        diags = []
        for sm in range(self.m+self.n-1):
            strip = []
            for (i, j) in filter(lambda square: sum(square)==sm, self.all_coords()):
                strip.append(self[i, j])
            diags.append(strip)
        return diags

    def antidiagonals(self) -> list[list[Piece]]:
        """Returns diagonals for checking win."""
        # Rotating board by 90 degree and then getting its diagonals
        rotated_board = Board(m=self.n, n=self.m)
        rotated_board.board = list(zip(*self.board))[::-1]
        return rotated_board.diagonals()

    def lines(self) -> list[list[Piece]]:
        """Returns all possible lines to be winned."""
        return self.board + self.columns() + self.diagonals() + self.antidiagonals()

    def flat(self) -> list[Piece]:
        """Returns flat format of board."""
        board = self.board
        return sum(board, [])
    
    def all_coords(self) -> list[tuple[int, int]]:
        """Returns all the board squares."""
        return [(i,j) for (i,j) in itertools.product(range(self.n), range(self.m))]

    def increase_dimensions(self, to_dim: int) -> None:
        """Scales up dimensions of infinite board to (to_dim, to_dim)."""
        newboard = Board(to_dim, to_dim)

        for (x, y) in newboard.all_coords():
            if (x, y) in self.all_coords():
                newboard[x, y] = self[x, y]
            else:
                newboard[x, y] = Piece()

        self.n, self.m = to_dim, to_dim
        self.board = newboard.board

    def __getitem__(self, key: tuple) -> None:
        if not key in self.all_coords() and not self.infinite:
            raise IndexError("Board index out of range.")

        if not key in self.all_coords() and self.infinite:
            to_dim = max(key[0]+1, key[1]+1)
            self.increase_dimensions(to_dim)

        return self.board[key[0]][key[1]]

    def __setitem__(self, key: tuple, value: Piece) -> None:
        if not key in self.all_coords() and not self.infinite:
            raise IndexError("Board index out of range.")

        if not key in self.all_coords() and self.infinite:
            to_dim = max(key[0]+1, key[1]+1)
            self.increase_dimensions(to_dim)
  
        self.board[key[0]][key[1]] = value

    def unicode_board(self) -> str:
        printed = []
        for row in self.board:
            for square in row:
                printed.append(square.unicode_symbol())
            printed.append("\n")
        return "".join(printed)

    def __repr__(self) -> str:
        printed = []
        for row in self.board:
            for square in row:
                printed.append(str(square))
            printed.append("\n")
        return "".join(printed)

    def get_attributes(self):
        '''Gets full information about game state for debugging purposes.'''
        attrs = []
        for key, value in self.__dict__.items():
           attrs.append(f'{key}: {value!r}\n')
        return ''.join(attrs)