from __future__ import annotations

import re

from pytactoe.board import Piece, Board


class TicTacToe(Board):
    """(m,n,k)-game."""

    def __init__(self, m: int = 3, n: int = 3, k: int = 3, infinite: bool = False, players: int = 2) -> None:
        """Initializes an (m,n,k)-game."""
        super().__init__(m, n, infinite)
        if k <= 0:
            raise ValueError("Parameter k must be positive.")
        self.k: int = k
        self.turn: Piece = Piece(1)
        self.players: int = players
        if not 2 <= players <= 8:
            raise ValueError("Number of players must be an int from 2 to 8.")
        self.ended: bool = False
        self.draw: bool = False
        self.winned: bool = False
        self.winner: Piece = Piece()
        self.move_count: int = 0
        self.fullmove_number: int = 1
        self.move_history: list[tuple[int, int]] = []

    def fen(self) -> str:
        """Creates the FEN representation of the game.

        For k-in-a-row games, we use FEN notation similar to one used in chess. 
        A FEN string contains three fields:
        1. Board representation. Rows from 0 to n-1 are separated by "/" symbol. Pieces are denoted 
        by its representation (i.e. "X", "O", etc.). A line of empty squares is denoted by its length.
        2. Active player. "X", "O", etc.
        3. The fullmove number. Starts with 1 and incremented after the last player's move ("O" if 
        there are only two players.)
        For example, the starting position of Tic Tac Toe is '3/3/3 X 1'.
        """
        fen = "/".join(("".join(map(repr, row)) for row in self.board))
        fen = re.sub(r"\.+", lambda x: str(len(x.group())), fen)
        fen = " ".join([fen, str(self.turn), str(self.fullmove_number)])
        return fen

    @classmethod
    def from_fen(cls, fen: str) -> TicTacToe:
        """Initializes a class instance from FEN. Does not check if given FEN is valid.
        Sets all parameters except m, n, turn, fullmove_number to default."""
        [position, turn, fullmove_number] = fen.split()

        position = re.sub(r"[0-9]+", lambda x: "."*int(x.group()), position)
        position = position.split("/")
        game = cls(m=len(position[0]), n=len(position))
        game.board = [list(map(Piece.from_symbol, row)) for row in position]
        
        game.turn = Piece.from_symbol(turn)
        game.fullmove_number = int(fullmove_number)
        
        return game

    def checkdraw(self) -> bool:
        """Checks if the game is a draw. Changes related attributes (draw, ended)."""
        if self.winned == True:
            return False
        if Piece() not in self.flat():
            self.ended = True
            self.draw = True
            return True
        return False

    def check_line_winned(line: list[Piece], k: int = 3) -> bool:
        """Checks if given line contain k-in-a-row."""
        if k > len(line):
            return False

        for i in range(len(line)-k+1):
            next_k_squares = line[i:i+k]
            if Piece() in next_k_squares:
                continue
            if all(square==next_k_squares[0] for square in next_k_squares):
                return True

        return False

    def checkwin(self) -> bool:
        """Checks if the game is winned by k-in-a-row. Changes related attributes (winner, winned, ended)."""
        winned = False
        for line in self.lines():
            winned = TicTacToe.check_line_winned(line, self.k)
            if winned:
                self.winner = self.turn
                self.winned = True
                self.ended = True
                break
        return winned
        
    def nextturn(self) -> None:
        """Changes turn to the next."""
        self.turn = self.turn.next(self.players)

    def revertturn(self) -> None:
        """Reverts turn."""
        self.turn = self.turn.previous(self.players)
    
    def is_legal_move(self, x: int, y: int) -> bool:
        """Checks move legality."""
        if not self.ended:
            try:
                return self[x, y] == Piece()
            except IndexError:
                return False
        return False

    def move(self, x: int, y: int) -> None:
        """Attempts to make a move. Checks for legality.
        If move attempt was succesful, calls checkwin(), checkdraw() and nextturn()."""
        if self.is_legal_move(x, y):
            self.fullmove_number += 1 if self.turn.code == self.players else 0
            self[x, y] = self.turn
            self.checkwin()
            self.checkdraw()
            self.nextturn()
            self.move_history.append((x,y))
            self.move_count += 1

    def pop(self) -> None:
        """Restores the position on the previous move. Does nothing if move history is empty."""
        if not self.move_count:
            return

        last_move = self.move_history[-1]

        self.draw = False
        self.ended = False
        self.winned = False
        self.winner = Piece()
        self.move_count -= 1
        self.fullmove_number -= 1 if self.turn == Piece(1) else 0
        self.move_history.pop()
        self[last_move[0], last_move[1]] = Piece()
        self.revertturn()

    def reset(self) -> None:
        """Resets the game to the default (empty) position. Saves parameters that were set at initialization."""
        self.board = [[Piece()]*self.m for _ in range(self.n)]
        self.turn = Piece(1)
        self.ended = False
        self.draw = False
        self.winned = False
        self.winner = Piece()
        self.move_count = 0
        self.fullmove_number = 1
        self.move_history = []

    def legalmoves(self) -> list[tuple[int, int]]:
        """Returns all possible moves."""
        return [(i,j) for (i,j) in self.all_coords() if self.is_legal_move(i,j)]