import itertools

from pytactoe.board import Piece, Board
from pytactoe.tictactoe import TicTacToe


class Connect(TicTacToe):
    """Game of Connect(m,n,k,p,q) family. Only implemented for two players."""

    def __init__(self, m: int = 19, n: int = 19, k: int = 6, p: int = 2, q: int = 1, infinite: bool = False) -> None:
        """Initializes a Connect(m,n,k,p,q) game."""
        super().__init__(m, n, k, infinite, players=2)
        self.p: int = p
        self.q: int = q

    def is_legal_move_sequence(self, moves: list[tuple[int, int]]):
        """Checks if move sequence is of valid length and checks its legality."""
        if self.move_count == 0 and len(moves) != self.q:
            return False
        if self.move_count > 0 and len(moves) != self.p:
            return False
        return all(self.is_legal_move(*move) for move in moves)

    def move_sequence(self, moves: list[tuple[int, int]]) -> None:
        """Attempts to make a move sequence (q regular moves if first move or p if not). Checks for legality.
        If move attempt was succesful, calls checkwin(), checkdraw() and nextturn()."""
        current_turn = self.turn
        if self.is_legal_move_sequence(moves):
            self.fullmove_number += 1 if self.turn.code == self.players else 0

            for move in moves:
                self[move[0], move[1]] = self.turn

                self.checkwin()
                self.checkdraw()
                
                self.move_history.append(move)
                self.move_count += 1

                if self.ended:
                    break
                self.turn = current_turn
            self.nextturn()

    def pop(self) -> None:
        if not self.move_count:
            return

        self.draw = False
        self.ended = False
        self.winned = False
        self.winner = Piece()

        self.move_count -= self.p
        self.fullmove_number -= 1 if self.turn.code == self.players else 0
        
        for _ in range(self.p):
            if not self.move_history:
                break
            last_move = self.move_history[-1]
            self.move_history.pop()
            self[last_move[0], last_move[1]] = Piece()

        self.revertturn()




class Pente(TicTacToe):
    """Pente game."""
    def __init__(self, m: int = 3, n: int = 3, k: int = 3, infinite: bool = False, players: int = 2, captures_to_win: int = 10) -> None:
        """Initializes a Pente game."""
        super().__init__(m, n, k, infinite, players)
        self.capture_count: list[int] = [0]*players
        self.captures_to_win: int = captures_to_win
        self.capture_history: list[list[tuple[int, int]]] = []

    def custodial_capture(self, x: int, y: int, length: int) -> int:
        """Does all possible custodial captures of given length on the given square.
        Returns the number of captured pieces."""
        capture_count = 0

        self.capture_history.append([])
        last_captures = self.capture_history[-1]

        for dir_x, dir_y in itertools.product((-1, 0, 1), repeat=2):
            if dir_x == 0 and dir_y == 0:
                continue

            try:
                if self[x, y] != self[x+(length+1)*dir_x, y+(length+1)*dir_y]:
                    continue
            except IndexError:
                continue
            
            if all(self[x+i*dir_x, y+i*dir_y] not in (self[x, y], Piece()) for i in range(1, length+1)):  
                for i in range(1, length+1):
                    self[x+i*dir_x, y+i*dir_y] = Piece()
                    last_captures.append((x+i*dir_x, y+i*dir_y))

            capture_count += length

        self.capture_count[self.turn.code-1] += capture_count
        return capture_count

    def check_win_by_captures(self) -> bool:
        """Checks if game is winned by captures. Changes related attributes (winner, winned, ended)."""
        if max(self.capture_count) >= self.captures_to_win:
            self.winner = self.turn
            self.winned = True
            self.ended = True
        return False

    def move(self, x: int, y: int) -> None:
        """Attempts to make a move. Checks for legality. Do possible custodial captures.
        If move attempt was succesful, calls checkwin(), check_win_by_captures(), checkdraw() and nextturn()."""
        if self.is_legal_move(x, y):
            self.fullmove_number += 1 if self.turn.code == self.players else 0
            self[x, y] = self.turn
            self.capture_count[self.turn.code-1] += self.custodial_capture(x, y, 2)
            self.checkwin()
            self.check_win_by_captures()
            self.checkdraw()
            self.nextturn()
            self.move_history.append((x,y))
            self.move_count += 1

    def pop(self) -> None:
        super().pop()
        for (i,j) in self.capture_history[-1]:
            self[i,j] = Piece()
        self.capture_history.pop()

    def reset(self) -> None:
        super().reset()
        self.capture_count: list[int] = [0]*self.players
        self.capture_history: list[list[tuple[int, int]]] = []


class UltimateTicTacToe(TicTacToe):
    """Ultimate (m,n,k)-game."""
    def __init__(self, m: int = 3, n: int = 3, k: int = 3, which_local_board: tuple[int, int] = (0, 0), players: int = 2) -> None:
        """Initializes an Ultimate (m,n,k)-game."""
        super().__init__(m, n, k, False, players)
        # Global board (self.board) contain winners of corresponding local boards.
        self.localboards: list[list[TicTacToe]] = [[TicTacToe(m, n, k, False, players) for _ in range(m)] for _ in range(n)]
        self.which_local_board: tuple[int, int] = which_local_board
        self.which_local_board_history: list[int] = [which_local_board]

    def is_legal_move(self, x: int, y: int) -> bool:
        """Checks move legality. Note that we can move on an already winned local board."""
        return self.localboards[self.which_local_board[0]][self.which_local_board[1]].is_legal_move(x, y)

    def move(self, x: int, y: int) -> None:
        """Attempts to make a move. Checks for legality. Chooses local board for next move.
        If move attempt was succesful, calls checkwin(), checkdraw() and nextturn() for both global and local boards."""
        if self.is_legal_move(x, y):
            self.fullmove_number += 1 if self.turn.code == self.players else 0

            localboard = self.localboards[self.which_local_board[0]][self.which_local_board[1]]
            localboard[x, y] = self.turn
            localboard.move_history.append((x,y))

            self.move_history.append((x,y))
            self.move_count += 1
            
            self.which_local_board = (x, y)
            self.which_local_board_history.append(self.which_local_board)

            if not localboard.winned:
                localboard.checkwin()
                localboard.checkdraw()

            self.board[x][y] = localboard.winner

            self.checkwin()
            self.checkdraw()
            self.nextturn()

    def pop(self) -> None:
        if not self.move_count:
            return

        self.which_local_board_history.pop()
        self.which_local_board = self.which_local_board_history[-1]
        
        localboard = self.localboards[self.which_local_board[0]][self.which_local_board[1]]
        localboard.pop()

        last_move = self.move_history[-1]

        self.draw = False
        self.ended = False
        self.winned = False
        self.winner = Piece()
        self.move_count -= 1
        self.fullmove_number -= 1 if self.turn == Piece(1) else 0
        self.move_history.pop()

        localboard.checkwin()
        localboard.checkdraw()

        self[last_move[0], last_move[1]] = localboard.winner
        
        self.revertturn()

    def reset(self) -> None:
        super().reset()
        self.localboards = [[TicTacToe(self.m, self.n, self.k, False, self.players) for _ in range(self.m)] for _ in range(self.n)]
        self.which_local_board = self.which_local_board_history[0]
        self.which_local_board_history = [self.which_local_board]

    def fen(self) -> str:
        pass

    @classmethod
    def from_fen(cls, fen: str) -> Board:
        pass



class MisereTicTacToe(TicTacToe):
    """Misere TicTacToe, which_local_board the player wins if the opponent "wins" in normal rules. """
    def __init__(self, m=3, n=3, k=3, infinite=False, players: int = 2) -> None:
        super().__init__(m, n, k, infinite, players)
    
    def checkwin(self) -> bool:
        if super().checkwin():
            self.winner = self.turn.previous(self.players)
            return True
        return False



class WildTicTacToe(TicTacToe):
    """ Tic Tac Toe which_local_board you can choose which piece to place. """
    def __init__(self, m=3, n=3, k=3, infinite=False, players: int = 2) -> None:
        super().__init__(m, n, k, infinite, players)
        self.move_history_pieces: list[Piece] = [] # Which pieces were chosen on each move

    def checkwin(self) -> bool:
        if super().checkwin():
            self.winner = Piece((self.move_count)%self.players+1)
        return self.winned

    def move(self, x: int, y: int, piece: Piece) -> None:
        if self.is_legal_move(x, y):
            self.fullmove_number += 1 if self.turn.code == self.players else 0
            self[x, y] = piece
            self.checkwin()
            self.checkdraw()
            self.nextturn()
            self.move_history.append((x,y))
            self.move_history_pieces.append(piece)
            self.move_count += 1

    def pop(self) -> None:
        super().pop()
        self.move_history_pieces.pop()



class ImpartialTicTacToe(WildTicTacToe):
    """ Tic Tac Toe with only one piece type. """
    def __init__(self, m=3, n=3, k=3, infinite=False, players: int = 2) -> None:
        super().__init__(m, n, k, infinite, players)

    def move(self, x: int, y: int) -> None:
        super().move(x, y, Piece(1))